#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# IDE: PyCharm 2017.2.4
# Author: Dajiang Ren<rendajiang@gagogroup.com>
# Created on 2018-06-13


from datetime import datetime

import boto3
from botocore.exceptions import ClientError

from . import StorageService, Object, Bucket, ObjectNotExistsError, ObjectExistsError


class S3Object(Object):
    def __init__(self, client, bucket_name: str, object_name: str, size: int, e_tag: str, last_modified: datetime):
        super().__init__(bucket_name, object_name, size, e_tag, last_modified)
        self.client = client

    def _download_to_file(self, local_file_path: str):
        self.client.download_file(Bucket=self.bucket_name, Key=self.object_name, Filename=local_file_path)

    def get_url(self):
        return '%s/%s/%s' % (self.client.meta.endpoint_url, self.bucket_name, self.object_name)


class S3Bucket(Bucket):
    def __init__(self, client, bucket_name: str):
        super().__init__(bucket_name)
        self.client = client

    def get_object(self, object_name: str) -> S3Object or None:
        if self.exist(object_name):
            header: dict = self.client.head_object(Bucket=self.bucket_name, Key=object_name)
            return S3Object(self.client, self.bucket_name, object_name, header['ContentLength'], header['ETag'],
                            header['LastModified'])
        else:
            raise ObjectNotExistsError

    def get_object_list(self, prefix: str=None) -> list:
        paginator = self.client.get_paginator('list_objects')
        objects: list = list()
        resume_token: str = None
        while True:
            if prefix is None:
                result = paginator.paginate(Bucket=self.bucket_name,
                                            PaginationConfig={'MaxItems': 1000,
                                                              'PageSize': 1000,
                                                              'StartingToken': resume_token})
            else:
                result = paginator.paginate(Bucket=self.bucket_name, Prefix=prefix,
                                            PaginationConfig={'MaxItems': 1000,
                                                              'PageSize': 1000,
                                                              'StartingToken': resume_token})
            for r in result:
                if 'Contents' in r:
                    contents: dict = r['Contents']
                else:
                    contents = list()
                if len(contents) == 0:
                    break
                for content in contents:
                    obj: S3Object = S3Object(self.client, self.bucket_name, content['Key'], content['Size'],
                                             content['ETag'], content['LastModified'])
                    objects.append(obj)
            resume_token = result.resume_token
            if not resume_token:
                break
        return objects

    def upload_file(self, local_file_path: str, object_name: str, acl: str='private',
                    overwrite_if_exists: bool=True) -> S3Object:
        if not self.exist(object_name) or overwrite_if_exists:
            self.client.upload_file(local_file_path, self.bucket_name, object_name)
        else:
            raise ObjectExistsError
        self.client.put_object_acl(Bucket=self.bucket_name, Key=object_name, ACL=acl)
        header: dict = self.client.head_object(Bucket=self.bucket_name, Key=object_name)
        return S3Object(self.client, self.bucket_name, object_name, header['ContentLength'], header['ETag'],
                        header['LastModified'])

    def upload_data(self, data: bytes, object_name: str, acl: str='private',overwrite_if_exists: bool=True) -> S3Object:
        if not self.exist(object_name) or overwrite_if_exists:
            self.client.put_object(ACL=acl, Body=data, Bucket=self.bucket_name, Key=object_name,
                                   ContentLength=len(data))
        else:
            raise ObjectExistsError
        header: dict = self.client.head_object(Bucket=self.bucket_name, Key=object_name)
        return S3Object(self.client, self.bucket_name, object_name, header['ContentLength'], header['ETag'],
                        header['LastModified'])

    def exist(self, object_name: str):
        try:
            self.client.head_object(Bucket=self.bucket_name, Key=object_name)
            return True
        except ClientError as error:
            if error.response['Error']['Code'] == '404':
                return False

    def remove(self, object_name: str):
        if self.exist(object_name):
            self.client.delete_object(Bucket=self.bucket_name, Key=object_name)

    def batch_remove(self, object_names: list):
        for i in range(0,len(object_names),1000):
            names: list = list()
            for o_n in object_names[i:i+1000]:
                names.append({'Key': o_n})
            self.client.delete_objects(Bucket=self.bucket_name, Delete={'Objects': names})

    def rename(self, object_name: str, new_object_name: str, acl: str='private'):
        if self.exist(object_name):
            copy_source = {
                'Bucket': self.bucket_name,
                'Key': object_name
            }
            self.client.copy(copy_source, Bucket=self.bucket_name, Key=new_object_name)
            self.client.put_object_acl(Bucket=self.bucket_name, Key=new_object_name, ACL=acl)
            self.remove(object_name)


class S3StorageService(StorageService):
    def __init__(self, auth_first: str, auth_second: str, region: str):
        super().__init__(auth_first, auth_second)
        self.client = boto3.client(
            's3',
            aws_access_key_id=auth_first,
            aws_secret_access_key=auth_second,
            region_name=region
        )

    def _get_bucket_impl(self, bucket_name: str) -> Bucket:
        return S3Bucket(self.client, bucket_name)


def create_service(first: str, second: str, region: str="cn-north-1"):
    return S3StorageService(first, second, region)


if __name__ == '__main__':
    s: S3StorageService = S3StorageService('AKIAPZKG6ANJSGGNTDSQ', 'UUNXCjb3KNiJp0rpEN9qDi02Oj73Xf5STLdaQKmo',
                                           'cn-northwest-1')
    b = s.exist('gagostore', 'abc')
    # s.download_object('gagostore', 'storage2.py', 'data/ss.text')
    # s.upload_file('gagostore', 'storage1.py', 'data/ss.text', 'public-read')
    os: list = s.get_object_list('gagostore', 'storage')
    dd = 1



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# IDE: PyCharm 2017.2.4
# Author: Dajiang Ren<rendajiang@gagogroup.com>
# Created on 2018-06-13

from importlib import import_module
from sys import modules
from datetime import datetime
import shutil
import os
from uuid import uuid4


class GagoStorageError(Exception):
    pass


class BucketNotExistsError(GagoStorageError):
    pass


class ObjectExistsError(GagoStorageError):
    pass


class ObjectNotExistsError(GagoStorageError):
    pass


class ACLError(GagoStorageError):
    pass


class Object:
    def __init__(self, bucket_name: str, object_name: str, size: int, e_tag: str, last_modified: datetime):
        self.bucket_name = bucket_name
        self.object_name = object_name
        self.size = size
        self.e_tag = e_tag
        self.last_modified = last_modified

    def download_to_file(self, local_file_path: str, overwrite_if_exists: bool=True):
        if not os.path.exists(local_file_path) or overwrite_if_exists:
            directory, file_name = os.path.split(local_file_path)
            temp_file_path = os.path.join(directory, str(uuid4()))
            if directory and not os.path.exists(directory):
                os.makedirs(directory, exist_ok=True)
            self._download_to_file(temp_file_path)
            if os.path.exists(local_file_path):
                os.remove(local_file_path)
            shutil.move(temp_file_path, local_file_path)
        else:
            raise FileExistsError

    def _download_to_file(self, local_file_path: str):
        pass

    def get_url(self):
        pass

    def get_gdal_url(self):
        return '/vsicurl/%s' % (self.get_url(),)


class Bucket:
    def __init__(self, bucket_name: str):
        self.bucket_name = bucket_name

    def get_object_list(self, prefix: str=None) -> list:
        pass

    def get_object(self, object_name: str) -> Object or None:
        pass

    def upload_file(self, local_file_path: str, object_name: str, acl: str, overwrite_if_exists: bool=True) -> Object:
        pass

    def upload_data(self, data: bytes, object_name: str, acl: str, overwrite_if_exists: bool=True) -> Object:
        pass

    def exist(self, object_name: str) -> bool:
        pass

    def remove(self, object_name: str):
        pass

    def batch_remove(self, object_names: list):
        pass

    def rename(self, object_name: str, new_object_name: str, acl: str='private'):
        pass


class StorageService:
    """
    a storage service provider abstract class
    """
    def __init__(self, auth_first: str, auth_second: str):
        self.auth_first = auth_first
        self.auth_second = auth_second
        self.bucket = None

    def get_bucket(self, bucket_name: str) -> Bucket:
        if self.bucket is None or self.bucket.bucket_name != bucket_name:
            self.bucket = self._get_bucket_impl(bucket_name)
        return self.bucket

    def _get_bucket_impl(self, bucket_name: str) -> Bucket:
        pass

    def get_object(self, bucket_name: str, object_name: str) -> Object or None:
        bucket: Bucket = self.get_bucket(bucket_name)
        return bucket.get_object(object_name)

    def get_object_list(self, bucket_name: str, prefix: str=None) -> list:
        bucket: Bucket = self.get_bucket(bucket_name)
        return bucket.get_object_list(prefix)

    def upload_file(self, bucket_name: str, object_name: str, local_file_path: str, acl: str='private',
                    overwrite_if_exists: bool=True) -> Object:
        """
        upload a file to object storage
        :param bucket_name: bucket name
        :param object_name: object name
        :param local_file_path: local file path
        :param acl: access control level, one of 'private'|'public-read'|'public-read-write'|'authenticated-read'
        :param overwrite_if_exists: indicates if overwrite the object when object already exists
        :return:
        """
        bucket: Bucket = self.get_bucket(bucket_name)
        return bucket.upload_file(local_file_path, object_name, acl, overwrite_if_exists)

    def upload_data(self, bucket_name, object_name: str, data: bytes, acl: str='private',
                    overwrite_if_exists: bool=True) -> Object:
        bucket: Bucket = self.get_bucket(bucket_name)
        return bucket.upload_data(data, object_name, acl, overwrite_if_exists)

    def download_object(self, bucket_name: str, object_name: str, local_file_path: str, overwrite_if_exists: bool=True):
        bucket: Bucket = self.get_bucket(bucket_name)
        obj: Object = bucket.get_object(object_name)
        obj.download_to_file(local_file_path, overwrite_if_exists)

    def exist(self, bucket_name: str, object_name: str) -> bool:
        bucket: Bucket = self.get_bucket(bucket_name)
        return bucket.exist(object_name)

    def remove(self, bucket_name: str, object_name: str):
        bucket: Bucket = self.get_bucket(bucket_name)
        bucket.remove(object_name)

    def batch_remove(self, bucket_name: str, object_names: list):
        bucket: Bucket = self.get_bucket(bucket_name)
        bucket.batch_remove(object_names)

    def rename(self, bucket_name: str, object_name: str, new_object_name: str, acl: str='private'):
        bucket: Bucket = self.get_bucket(bucket_name)
        bucket.rename(object_name, new_object_name, acl)


def create_service(provider_name: str, auth_first: str, auth_second: str, region: str) -> StorageService:
    """
    create a service
    :param provider_name: s3, azure, aliyun, cephrgw
    :param auth_first: first param
    :param auth_second: second param
    :param region: region name for s3, oss or end point url for Azure Blob
    :return: StorageService
    """
    if __package__ is None:
        module_name = provider_name
        import_module(module_name)
        m = modules[module_name]
    else:
        module_name: str = '.' + provider_name
        import_module(module_name, package=__package__)
        m = modules[__package__ + "." + provider_name]
    if hasattr(m, 'create_service'):
        return m.create_service(auth_first, auth_second, region)


if __name__ == '__main__':
    service: StorageService = create_service('s3', 'AKIAPZKG6ANJSGGNTDSQ', 'UUNXCjb3KNiJp0rpEN9qDi02Oj73Xf5STLdaQKmo',
                                             'cn-north-1')
    dd = 1



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# IDE: PyCharm 2017.2.4
# Author: Dajiang Ren<rendajiang@gagogroup.com>
# Created on 2018-09-26

import os
import shutil
import datetime

from . import StorageService, Bucket, Object, ObjectNotExistsError, ObjectExistsError


class FileSystemObject(Object):
    def __init__(self, bucket_name: str, object_name: str, size: int, e_tag: str, last_modified: datetime.datetime,
                 directory: str):
        super().__init__(bucket_name, object_name, size, e_tag, last_modified)
        self.full_path = os.path.join(directory, object_name)

    def _download_to_file(self, local_file_path: str):
        local_dir: str = os.path.dirname(local_file_path)
        if local_dir and not os.path.exists(local_dir):
            os.makedirs(local_dir, exist_ok=True)
        shutil.copy(self.full_path, local_file_path)

    def get_url(self):
        return self.full_path

    def get_gdal_url(self):
        return self.full_path


class FileSystemBucket(Bucket):
    def __init__(self, bucket_name: str, root: str):
        super().__init__(bucket_name)
        self.directory = os.path.join(root, bucket_name)
        if not os.path.exists(self.directory):
            os.makedirs(self.directory, exist_ok=True)

    def get_object_list(self, prefix: str=None) -> list:
        file_list = list()
        files = os.listdir(self.directory)
        for f in files:
            full_path: str = os.path.join(self.directory, f)
            if os.path.isdir(full_path):
                pass
            elif os.path.isfile(full_path):
                file_list.append(f)
        return files

    def get_object(self, object_name: str) -> Object or None:
        if self.exist(object_name):
            full_path = os.path.join(self.directory, object_name)
            return FileSystemObject(self.bucket_name, object_name, os.path.getsize(full_path), '',
                                    os.path.getmtime(full_path), self.directory)

    def upload_file(self, local_file_path: str, object_name: str, acl: str, overwrite_if_exists: bool=True) -> Object:
        if not os.path.exists(local_file_path):
            raise FileNotFoundError
        if self.exist(object_name):
            if overwrite_if_exists:
                self.remove(object_name)
            else:
                raise ObjectExistsError
        dst_path = os.path.join(self.directory, object_name)
        dir_name = os.path.dirname(dst_path)
        os.makedirs(dir_name, exist_ok=True)
        shutil.copy(local_file_path, dst_path)
        return self.get_object(object_name)

    def upload_data(self, data: bytes, object_name: str, acl: str, overwrite_if_exists: bool=True) -> Object:
        if self.exist(object_name):
            if overwrite_if_exists:
                self.remove(object_name)
            else:
                raise ObjectExistsError
        dst_path = os.path.join(self.directory, object_name)
        dir_name = os.path.dirname(dst_path)
        os.makedirs(dir_name, exist_ok=True)
        with open(dst_path, 'wb') as fp:
            fp.write(data)
        return self.get_object(object_name)

    def exist(self, object_name: str) -> bool:
        full_path = os.path.join(self.directory, object_name)
        return os.path.exists(full_path)

    def remove(self, object_name: str):
        if self.exist(object_name):
            full_path = os.path.join(self.directory, object_name)
            os.remove(full_path)
        else:
            raise ObjectNotExistsError

    def batch_remove(self, object_names: list):
        for o_n in object_names:
            self.remove(o_n)

    def rename(self, object_name: str, new_object_name: str, acl: str='private'):
        if self.exist(object_name):
            src_path: str = os.path.join(self.directory, object_name)
            dst_path: str = os.path.join(self.directory, new_object_name)
            shutil.move(src_path, dst_path)
        else:
            raise ObjectNotExistsError


class FileSystemStorageService(StorageService):
    def __init__(self, first: str, second: str, region: str=''):
        super().__init__(first, second)
        self.root = region

    def _get_bucket_impl(self, bucket_name: str):
        return FileSystemBucket(bucket_name, self.root)


def create_service(first: str, second: str, region: str=''):
    return FileSystemStorageService(first, second, region)


if __name__ == '__main__':
    f: StorageService = FileSystemStorageService('', '')
    b: Bucket = f.get_bucket('123')
    b.upload_file('s3.py', '123.py', acl='public', overwrite_if_exists=True)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# IDE: PyCharm 2017.2.4
# Author: Dajiang Ren<rendajiang@gagogroup.com>
# Created on 2018-09-26


from distutils.core import setup #如果没有需要先安装


setup(
    name='gagoos',  #打包后的包文件名
    version='0.1.1',  #版本
    description='Gago Object Storage',
    author='rockdj1983',
    author_email='rockdj1983@163.com',
    url='http://www.gagogroup.cn',
    packages=['gagoos'],  #与前面的新建文件名一致
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# IDE: PyCharm 2017.2.4
# Author: Dajiang Ren<rendajiang@gagogroup.com>
# Created on 2018-06-13


from .storage import StorageService, Bucket, Object, ObjectNotExistsError, ObjectExistsError, ACLError



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# IDE: PyCharm 2017.2.4
# Author: Dajiang Ren<rendajiang@gagogroup.com>
# Created on 2018-06-14


from datetime import datetime

import oss2

from . import StorageService, Bucket, Object, ObjectNotExistsError, ObjectExistsError, ACLError

from urllib.parse import urlparse

acl_mapping = {
    'private': oss2.OBJECT_ACL_PRIVATE,
    'public-read': oss2.OBJECT_ACL_PUBLIC_READ,
    'public-read-write': oss2.OBJECT_ACL_PUBLIC_READ_WRITE,
    'authenticated-read': oss2.OBJECT_ACL_PRIVATE
}


class OSSObject(Object):
    def __init__(self, bucket: oss2.Bucket, object_name: str, size: int, e_tag: str, last_modified: datetime):
        super().__init__(bucket.bucket_name, object_name, size, e_tag, last_modified)
        self.bucket: oss2.Bucket = bucket

    def _download_to_file(self, local_file_path: str):
        self.bucket.get_object_to_file(self.object_name, local_file_path)

    def get_url(self):
        url_split = urlparse(self.bucket.endpoint)
        return 'https://%s.%s/%s' % (self.bucket_name, url_split[1], self.object_name)


class OSSBucket(Bucket):
    def __init__(self, auth: oss2.Auth, region: str, bucket_name: str):
        super().__init__(bucket_name)
        self.bucket: oss2.Bucket = oss2.Bucket(auth, region, bucket_name)

    def get_object(self, object_name: str) -> OSSObject:
        if self.exist(object_name):
            try:
                meta: oss2.models.GetObjectMetaResult = self.bucket.get_object_meta(object_name)
            except oss2.exceptions.NoSuchKey:
                raise ObjectNotExistsError
            return OSSObject(self.bucket, object_name, meta.content_length, meta.etag,
                             datetime.fromtimestamp(meta.last_modified))
        else:
            raise ObjectNotExistsError

    def get_object_list(self, prefix: str=None) -> list:
        marker: str = ''
        objects: list = list()
        while True:
            results: oss2.models.ListObjectsResult = self.bucket.list_objects(prefix=prefix, marker=marker,
                                                                              max_keys=1000)
            for r in results.object_list:
                object_info: oss2.models.SimplifiedObjectInfo = r
                objects.append(OSSObject(self.bucket, object_info.key, object_info.size, object_info.etag,
                                         datetime.fromtimestamp(object_info.last_modified)))
            marker = results.next_marker
            if not marker:
                break
        return objects

    def upload_file(self, local_file_path: str, object_name: str, acl: str, overwrite_if_exists: bool=True):
        if not self.exist(object_name) or overwrite_if_exists:
            if acl not in acl_mapping:
                raise ACLError
            self.bucket.put_object_from_file(object_name, local_file_path)
            self.bucket.put_object_acl(object_name, acl_mapping[acl])
            return self.get_object(object_name)
        else:
            raise ObjectExistsError

    def upload_data(self, data: bytes, object_name: str, acl: str, overwrite_if_exists: bool=True):
        if not self.exist(object_name) or overwrite_if_exists:
            if acl not in acl_mapping:
                raise ACLError
            self.bucket.put_object(object_name, data)
            self.bucket.put_object_acl(object_name, acl_mapping[acl])
            return self.get_object(object_name)
        else:
            raise ObjectExistsError

    def exist(self, object_name: str):
        return self.bucket.object_exists(object_name)

    def remove(self, object_name: str):
        if self.exist(object_name):
            self.bucket.delete_object(object_name)

    def batch_remove(self, object_names: list):
        for i in range(0, len(object_names), 1000):
            names: list = object_names[i:i+1000]
            self.bucket.batch_delete_objects(names)

    def rename(self, object_name: str, new_object_name: str, acl: str='private'):
        if self.exist(object_name):
            self.bucket.copy_object(self.bucket_name, object_name, new_object_name)
            self.bucket.put_object_acl(new_object_name, acl)
            self.remove(object_name)


class OSSStorageService(StorageService):
    def __init__(self, auth_first: str, auth_second: str, region: str):
        super().__init__(auth_first, auth_second)
        self.auth = oss2.Auth(auth_first, auth_second)
        self.service = oss2.Service(self.auth, region)

    def _get_bucket_impl(self, bucket_name: str) -> OSSBucket:
        return OSSBucket(self.auth, self.service.endpoint, bucket_name)


def create_service(first: str, second: str, region: str="https://oss-cn-hangzhou.aliyuncs.com"):
    return OSSStorageService(first, second, region)

